import re
import gzip
import shutil
from os import chdir
from ftplib import FTP
from pathlib import Path

from Bio import SeqIO
from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq
import ncbi.datasets

from metaerg import context
from metaerg.datatypes import fasta
from metaerg.datatypes import sqlite
from metaerg.datatypes.blast import BlastResult, DBentry, TabularBlastParser
from metaerg.datatypes import ncbi_ftp

DB_DESCRIPTIONS_FILENAME = 'db_descriptions.txt'
DB_TAXONOMY_FILENAME = 'db_taxonomy.txt'
DB_PROTEINS_FILENAME = 'db_protein.faa'
DB_RNA_FILENAME = 'db_rna.fna'

RELEVANT_FEATURE_TYPES = 'CDS rRNA tRNA RNase_P_RNA SRP_RNA riboswitch snoRNA ncRNA tmRNA antisense_RNA binding_site ' \
                         'hammerhead_ribozyme scRNA mobile_element mobile_genetic_element misc_RNA ' \
                         'autocatalytically_spliced_intron'.split()
IGNORED_FEATURE_TYPES = 'gene pseudogene exon direct_repeat region sequence_feature pseudogenic_tRNA pseudogenic_rRNA ' \
                        'repeat_region ribosome_entry_site minus_10_signal minus_35_signal protein_binding_site ' \
                        'regulatory transcript mRNA gap assembly_gap source misc_feature variation misc_structure ' \
                        'transcript 3\'UTR 5\'UTR intron signal_peptide_region_of_CDS sequence_alteration'.split()


def _run_programs(genome_name, contig_dict, db_connection, result_files):
    rna_nt_file = context.spawn_file('rna.fna', genome_name)
    blastn_result_file = context.spawn_file('blastn', genome_name)
    if rna_nt_file.exists() and rna_nt_file.stat().st_size:
        if context.FORCE or not blastn_result_file.exists() or not blastn_result_file.stat().st_size:
            blastn_db = Path(context.DATABASE_DIR, DB_RNA_FILENAME)
            context.run_external(f'blastn -db {blastn_db} -query {rna_nt_file} -out {blastn_result_file} -max_target_seqs 10 '
                                 f'-outfmt 6')
    else:
        context.log(f'({genome_name}) Skipping blastn, fasta file with RNA genes missing or empty.')

    cds_aa_file = context.spawn_file('cds.faa', genome_name)
    diamond_db = Path(context.DATABASE_DIR, DB_PROTEINS_FILENAME)
    context.run_external(f'diamond blastp -d {diamond_db} -q {cds_aa_file} -o {result_files[0]} -f 6 '
                         f'--threads {context.CPUS_PER_GENOME} --fast --max-target-seqs 10')


def _read_results(genome_name, contig_dict, db_connection, result_files) -> int:
    # (1) load databse descriptions
    db_descr = Path(context.DATABASE_DIR, DB_DESCRIPTIONS_FILENAME)
    descriptions = {'p': {}, 'e': {}, 'v': {}}
    entries_without_descr = 0
    with open(db_descr) as descr_handle:
        for line in descr_handle:
            words = line.split('\t')
            descriptions[words[0]][int(words[1])] = words[2].strip()
            if not descriptions[words[0]][int(words[1])]:
                entries_without_descr += 1
    context.log(f'({genome_name}) Parsed ({len(descriptions["p"])}, {len(descriptions["e"])}, {len(descriptions["v"])})'
                f' gene descriptions from db for (prokaryotes, eukaryotes and viruses) respectively. Empty'
                f' descriptions: {entries_without_descr}.')
    # (2) load database taxonomy
    db_taxonomy = Path(context.DATABASE_DIR, DB_TAXONOMY_FILENAME)
    taxonomy = {'p': {}, 'e': {}, 'v': {}}
    with open(db_taxonomy) as taxon_handle:
        for line in taxon_handle:
            words = line.split('\t')
            taxonomy[words[0]][int(words[1])] = words[2].strip().replace('~', '; ')
    context.log(f'({genome_name}) Parsed ({len(taxonomy["p"])}, {len(taxonomy["e"])}, {len(taxonomy["v"])}) '
                f'taxa from db for (prokaryotes, eukaryotes and viruses) respectively.')
    # (3) parse diamond blast results

    def process_blast_result(blast_result: BlastResult):
        feature = sqlite.read_feature_by_id(db_connection, blast_result.query())
        if not feature:
            raise Exception(f'Found results for unknown feature {blast_result.query()}, '
                            f'may need to rerun metaerg with --force')
        feature.blast = blast_result
        feature.descr = blast_result.hits[0].hit.descr
        feature.taxon =  blast_result.hits[0].hit.taxon
        sqlite.update_feature_in_db(db_connection, feature)

    def dbentry_from_string(db_id: str) -> DBentry:
        w = db_id.split('~')
        return DBentry(domain=w[0], taxon=taxonomy[w[0]][int(w[1])], descr=descriptions[w[0]][int(w[2])],
                       accession=w[3], gene=w[4], length=int(w[5]), pos=int(w[6]))

    blast_result_count = 0
    with TabularBlastParser(result_files[0], 'BLAST', dbentry_from_string) as handle:
        for blast_result in handle:
            blast_result_count += 1
            process_blast_result(blast_result)

    blastn_result_file = context.spawn_file('blastn', genome_name)
    with TabularBlastParser(blastn_result_file, 'BLAST', dbentry_from_string) as handle:
        for blast_result in handle:
            blast_result_count += 1
            process_blast_result(blast_result)
    return blast_result_count


@context.register_annotator
def run_and_read_diamond_blastn():
    return ({'pipeline_position': 81,
             'annotator_key': 'diamond_and_blastn',
             'purpose': 'function prediction and taxonomic classification of genes with diamond and blastn',
             'programs': ('diamond', 'blastn'),
             'databases': (Path(DB_PROTEINS_FILENAME + '.dmnd'),
                           Path(DB_RNA_FILENAME + '.nhr'),
                           Path(DB_DESCRIPTIONS_FILENAME),
                           Path(DB_TAXONOMY_FILENAME)),
             'result_files': ('diamond',),
             'run': _run_programs,
             'read': _read_results})


def init_pristine_db_dir(dir) -> tuple[Path, Path, Path, Path]:
    fasta_protein_db = Path(dir, DB_PROTEINS_FILENAME)
    fasta_nt_db = Path(dir, DB_RNA_FILENAME)
    descr_db = Path(dir, DB_DESCRIPTIONS_FILENAME)
    taxon_db = Path(dir, DB_TAXONOMY_FILENAME)
    fasta_protein_db.unlink(missing_ok=True)
    fasta_nt_db.unlink(missing_ok=True)
    descr_db.unlink(missing_ok=True)
    taxon_db.unlink(missing_ok=True)
    return fasta_protein_db, fasta_nt_db, descr_db, taxon_db


def update_db_descriptions_get_db_id(description, dictionary, file_handle, kingdom):
    try:
        descr_id = dictionary[description]
    except KeyError:
        descr_id = len(dictionary)
        dictionary[description] = descr_id
        file_handle.write(f'{kingdom}\t{descr_id}\t{description}\n')
    return descr_id


@context.register_database_installer
def install_viral_database():
    if 'V' not in context.TASKS:
        return
    context.log('Downloading viral refseq from the NCBI...')
    VIR_DB_DIR = Path(context.DATABASE_DIR, 'ncbi-cache', 'vir')
    VIR_DB_DIR.mkdir(exist_ok=True, parents=True)
    if Path(VIR_DB_DIR, DB_DESCRIPTIONS_FILENAME).exists() and not context.FORCE:
        context.log(f'Keeping existing viral database in {VIR_DB_DIR}, use --force to overwrite.')
        return
    fasta_protein_db, fasta_nt_db, descr_db, taxon_db = init_pristine_db_dir(VIR_DB_DIR)

    descr_dict = dict()
    taxon_dict = dict()

    pattern = re.compile(r'\s*\[(.+?)]$')  # \]
    with open(fasta_protein_db, 'w') as prot_fasta_out_handle, \
            open(descr_db, 'w') as descr_handle,\
            open(taxon_db, 'w') as taxon_handle:
        for i in range(1,5):
            f = Path(VIR_DB_DIR, f'viral.{i}.protein.gpff.gz')
            if not f.exists() or not f.stat().st_size:
                url = 'ftp.ncbi.nlm.nih.gov/refseq/release/viral/viral'
                context.run_external(f'wget -P {VIR_DB_DIR} {url}.{i}.protein.gpff.gz')
            gene_count = 0
            with gzip.open(f, 'rt') as file_handle:
                for gb_record in SeqIO.parse(file_handle, "genbank"):
                    if gb_record.annotations['molecule_type'] != 'protein':
                        context.log("warning: skipping non-coding gene in viral refseq")
                        continue
                    match = pattern.search(gb_record.description)
                    if match:
                        gb_record.annotations['taxonomy'].append(match.group(1))
                        gb_record.description = gb_record.description[0:match.start()]
                    else:
                        context.log(f'Warning, Failed to parse viral species from {gb_record.description}')
                    taxon_str = "~".join(gb_record.annotations['taxonomy'])
                    try:
                        taxon_id = taxon_dict[taxon_str]
                    except KeyError:
                        taxon_id = len(taxon_dict)
                        taxon_dict[taxon_str] = taxon_id
                        taxon_handle.write(f'v\t{taxon_id}\t{taxon_str}\n')
                    descr_id = update_db_descriptions_get_db_id(gb_record.description, descr_dict, descr_handle, 'v')
                    seq_record = SeqRecord(gb_record.seq)
                    seq_record.id = '{}~{}~{}~{}~{}~{}~{}'.format('v', taxon_id, descr_id,
                                                                  gb_record.id.replace('~', '-'), '',
                                                                  len(seq_record.seq), gene_count)
                    seq_record.description = gb_record.description
                    SeqIO.write(seq_record, prot_fasta_out_handle, "fasta")
                    gene_count += 1


@context.register_database_installer
def install_eukaryote_database():
    if 'E' not in context.TASKS:
        return
    context.log('Downloading taxon-unique eukaryotic genomes using NCBI Datasets...')
    EUK_DB_DIR = Path(context.DATABASE_DIR, 'ncbi-cache', 'euk')
    EUK_DB_DIR.mkdir(exist_ok=True, parents=True)
    if Path(EUK_DB_DIR, DB_DESCRIPTIONS_FILENAME).exists() and not context.FORCE:
        context.log(f'Keeping existing eukaryote database in {EUK_DB_DIR}, use --force to overwrite.')
        return
    fasta_protein_db, fasta_nt_db, descr_db, taxon_db = init_pristine_db_dir(EUK_DB_DIR)
    chdir(context.DATABASE_DIR)
    # (1) probe the NCBI for available eukaryotic assemblies
    api_instance = ncbi.datasets.GenomeApi(ncbi.datasets.ApiClient())
    ASSEMBLY_LEVELS = 'Scaffold', 'Chromosome', 'Complete Genome'
    targets = dict()
    taxa_done = dict()
    for tax_name in 'Amoebozoa Ancyromonadida Apusozoa Breviatea CRuMs Cryptophyceae Discoba Glaucocystophyceae ' \
                'Haptista Hemimastigophora Malawimonadida Metamonada Rhodelphea Rhodophyta Sar Aphelida ' \
                'Choanoflagellata Filasterea Fungi Ichthyosporea Rotosphaerida'.split():
        genome_summary = api_instance.assembly_descriptors_by_taxon(taxon=tax_name, page_size=1000)
        count = 0
        if genome_summary.assemblies:
            for assembly in map(lambda d: d.assembly, genome_summary.assemblies):
                if assembly.assembly_accession[:3] == 'GCF':
                    level_of_new_assembly = assembly.assembly_level
                    if assembly.org.tax_id in taxa_done.keys():
                        if level_of_new_assembly in ASSEMBLY_LEVELS and ASSEMBLY_LEVELS.index(level_of_new_assembly) \
                                > taxa_done[assembly.org.tax_id]:
                            taxa_done[assembly.org.tax_id] = ASSEMBLY_LEVELS.index(level_of_new_assembly)
                            targets[assembly.org.tax_id] = assembly.assembly_accession.strip()
                        else:
                            continue
                    else:
                        count += 1
                        taxa_done[assembly.org.tax_id] = ASSEMBLY_LEVELS.index(level_of_new_assembly) \
                            if level_of_new_assembly in ASSEMBLY_LEVELS else 0
                        targets[assembly.org.tax_id] = assembly.assembly_accession.strip()
        try:
            total_count = int(genome_summary.total_count)
        except TypeError:
            total_count = 0
        context.log(f"({count}/{total_count}) taxon-unique assemblies for {tax_name}")
    total_target_count = len(targets)
    context.log(f'Found  {total_target_count} targets for download in total.')
    # (2) Download new assemblies not yet in local cache
    targets = [accession for accession in targets.values() if not Path(EUK_DB_DIR, f'{accession}.gbk.gz').exists()]
    context.log(f'Of those, {total_target_count - len(targets)} already in local cache. ')

    api_response = api_instance.download_assembly_package(targets, exclude_sequence = False,
                                                          include_annotation_type = ['GENOME_GBFF'],
                                                          _preload_content = False)
    zipfile = Path(context.DATABASE_DIR, 'ncbi_genomes.zip')
    zipfile.unlink(missing_ok=True)
    with open(zipfile, 'wb') as f:
        f.write(api_response.data)
    if not zipfile.exists():
        context.log('No success in downloading eukaryote genomes, aborting...')
        return
    context.run_external(f'unzip -qq -o -d {context.DATABASE_DIR} {zipfile}', log_cmd=False)
    # (3) move gbk file into local cache gzipped
    count=0
    for accession in targets:
        count += 1
        src_dir = Path(context.DATABASE_DIR, 'ncbi_dataset', 'data', accession)
        dest_file = Path(EUK_DB_DIR, f'{accession}.gbk.gz')
        if dest_file.exists():
            continue
        context.log(f'({count}/{len(targets)}) Now extracting {accession} into ncbi-cache as {dest_file}...')
        gbff_count = 0
        for file in src_dir.glob('*.gbff'):
            with open(Path(src_dir, file), 'rb') as f_in, gzip.open(dest_file, 'wb') as f_out:
                f_out.writelines(f_in)
            gbff_count += 1
        if gbff_count > 1:
            context.log(f'WARNING: {src_dir} has {gbff_count} gbff files')
        elif gbff_count == 0:
            context.log(f'WARNING: {src_dir} has no gbff files')
    for file in zipfile, Path(context.DATABASE_DIR, '../../README.md'):
        file.unlink(missing_ok=True)
    # (4) build blast databases
    descr_dict = dict()
    taxon_dict = dict()
    with open(fasta_protein_db, 'w') as prot_fasta_out_handle,open(fasta_nt_db, 'w') as rna_fasta_out_handle, \
            open(descr_db, 'w') as descr_handle, open(taxon_db, 'w') as taxon_handle:
        genome_count = 0
        for file in EUK_DB_DIR.glob('*.gbk.gz'):
            genome_count += 1
            context.log(f'({genome_count}/{len(targets)}) Now extracting "{file}"')
            file_path = Path(EUK_DB_DIR, file)
            feature_success_counter = 0
            feature_total_counter = 0
            without_translation_counter = 0
            with gzip.open(file_path, "rt") as handle:
                first = True
                for gb_record in SeqIO.parse(handle, "genbank"):
                    if first:
                        taxon = gb_record.annotations['taxonomy']
                        taxon.append(gb_record.annotations['organism'])
                        taxon_str = "~".join(taxon)
                        try:
                            taxon_id = taxon_dict[taxon_str]
                        except KeyError:
                            taxon_id = len(taxon_dict)
                            taxon_dict[taxon_str] = taxon_id
                            taxon_handle.write(f'e\t{taxon_id}\t{taxon_str}\n')
                        first = False
                    feature_total_counter += len(gb_record.features)
                    for feature in gb_record.features:
                        if feature.type in RELEVANT_FEATURE_TYPES:
                            if feature.type == 'CDS':
                                if not 'translation' in feature.qualifiers:
                                    without_translation_counter += 1
                                    continue
                                seq_record = SeqRecord(Seq(feature.qualifiers['translation'][0]))
                                file_handle = prot_fasta_out_handle
                            else:
                                seq_record = SeqRecord(feature.extract(gb_record.seq))
                                file_handle = rna_fasta_out_handle
                            if 'product' in feature.qualifiers:
                                seq_record.description = feature.qualifiers['product'][0]
                            elif 'gene' in feature.qualifiers:
                                seq_record.description = feature.qualifiers['gene'][0]
                            elif 'note' in feature.qualifiers:
                                seq_record.description = feature.qualifiers['note'][0]
                            elif 'mobile_element_type' in feature.qualifiers:
                                seq_record.description = feature.qualifiers['mobile_element_type'][0]
                            elif 'ncRNA_class' in feature.qualifiers:
                                seq_record.description = feature.qualifiers['ncRNA_class'][0]
                            else:
                                context.log(f"WARNING: No description for feature {feature}")
                                seq_record.description = f'Unknown {feature.type}'
                            descr_id = update_db_descriptions_get_db_id(seq_record.description, descr_dict,
                                                                        descr_handle, 'e')
                            seq_record.id = '{}~{}~{}~{}~{}~{}~{}'.format('e', taxon_id, descr_id,
                                                                          gb_record.id.replace('~', '-'), '',
                                                                          len(seq_record.seq), feature_success_counter)
                            SeqIO.write(seq_record, file_handle, "fasta")
                            feature_success_counter += 1
                        elif feature.type in IGNORED_FEATURE_TYPES:
                            pass
                        else:
                            context.log(f'  Warning: unknown feature "{feature}"')
            context.log(f'  ... extracted ({feature_success_counter}/{feature_total_counter}) proteins and RNAs. '
                        f'{without_translation_counter} proteins skipped for lack of translation.')


@context.register_database_installer
def install_prokaryote_database():
    if 'P' not in context.TASKS:
        return
    context.log('Downloading a gff file from the NCBI FTP server for each prokaryotic genome in gtdbtk...')
    PROK_DB_DIR = Path(context.DATABASE_DIR, 'ncbi-cache', 'pro')
    PROK_DB_DIR.mkdir(exist_ok=True, parents=True)
    PROK_DB_DIR_FAA = Path(PROK_DB_DIR, 'faa')
    PROK_DB_DIR_FAA.mkdir(exist_ok=True)
    PROK_DB_DIR_FNA = Path(PROK_DB_DIR, 'fna')
    PROK_DB_DIR_FNA.mkdir(exist_ok=True)
    if Path(PROK_DB_DIR, DB_DESCRIPTIONS_FILENAME).exists() and not context.FORCE:
        context.log(f'Keeping existing prokaryote database in {PROK_DB_DIR}, use --force to overwrite.')
        return
    RNA_DESCR_RE = re.compile(r'\[product=(.+?)]')
    AA_DESCR_RE = re.compile(r'\[(.+?)]$')
    fasta_protein_db, fasta_nt_db, descr_db, taxon_db = init_pristine_db_dir(PROK_DB_DIR)
    descr_dict = {}
    kingdom = 'p'
    # (2) Load and prep list of target taxa from GTDB
    taxa_count = 0
    gtdbtk_taxonomy_file = Path(context.GTDBTK_DIR, 'taxonomy', 'gtdb_taxonomy.tsv')
    # determine line count
    with open(gtdbtk_taxonomy_file) as taxonomy_handle:
        for line in taxonomy_handle:
            taxa_count += 1
    taxon_list = []
    context.log(f'GTDBTK comprises {taxa_count} taxa...')
    with open(gtdbtk_taxonomy_file) as taxonomy_handle, open(descr_db, 'w') as descr_handle:
        success_count = 0
        genomes_in_cache_count = 0
        count = 0
        for line in taxonomy_handle:
            count += 1
            words = line.split()
            accession = words[0][3:]
            taxonomy = re.sub("\w__", "", words[1]).split(';')
            future_faa_file = Path(PROK_DB_DIR_FAA, f'{accession}.faa.gz')
            future_rna_file = Path(PROK_DB_DIR_FNA, f'{accession}.rna.fna.gz')
            taxon = {'id': len(taxon_list), 'accession': accession, 'taxonomy': taxonomy, 'in_local_cache': False}
            taxon_list.append(taxon)
            if future_faa_file.exists() and future_rna_file.exists():
                taxon['in_local_cache'] = True
                download_status = '++'
                genomes_in_cache_count += 1
            else:
                try:
                    download_status = ''
                    targets = []
                    if not future_faa_file.exists():
                        targets.append(('_protein.faa.gz', future_faa_file))
                    if not future_rna_file.exists():
                        targets.append(('_rna_from_genomic.fna.gz', future_rna_file))
                    success = ncbi_ftp.fetch(accession, targets)
                    if success[0]:
                        download_status += '*'
                        success_count += 1
                        taxon['in_local_cache'] = True
                        genomes_in_cache_count += 1
                    if success[1]:
                        download_status += '*'
                    if not download_status:
                        continue
                except EOFError:
                    context.log('FTP Error at NCBI - resetting connection')
                    ftp = FTP('ftp.ncbi.nlm.nih.gov')
                    ftp.login()
                    continue
                except BrokenPipeError:
                    context.log('FTP Error at NCBI - resetting connection')
                    ftp = FTP('ftp.ncbi.nlm.nih.gov')
                    ftp.login()
                    continue
            context.log(f'({count}/{taxa_count}) {download_status} {future_faa_file.name} {taxonomy}')
            # extract_proteins_and_rna_prok(taxon, descr_dict, PROK_DB_DIR)
            for input, output in ((future_faa_file, fasta_protein_db), (future_rna_file, fasta_nt_db)):
                if not input.exists():
                    continue
                with fasta.FastaParser(input, cleanup_seq=False) as reader, open(output, 'a') as writer:
                    gene_counter = 0
                    for f in reader:
                        if m:= RNA_DESCR_RE.search(f.descr):
                            descr = m.group(1)
                        elif m:= AA_DESCR_RE.search(f.descr):
                            descr = f.descr[0:m.start()].strip()
                            descr = descr.replace('MULTISPECIES: ', '')
                        descr_id = update_db_descriptions_get_db_id(descr, descr_dict, descr_handle, kingdom)
                        f.id = '{}~{}~{}~{}~{}~{}~{}'.format(kingdom, taxon['id'], descr_id, f.id.replace('~', '-'),
                                                             '', len(f), gene_counter)
                        f.descr = descr
                        gene_counter += 1
                        fasta.write_fasta(writer, f)

    context.log(f'downloaded {int(success_count)} new genomes. Total prok genomes in cache: {genomes_in_cache_count}.')
    # (4) save taxonomy (includes status)
    context.log(f'Now writing taxonomy file for prok.')
    with open(taxon_db, "w") as taxon_handle:
        for t in taxon_list:
            taxon_handle.write(f'p\t{t["id"]}\t{"~".join(t["taxonomy"])}\t{t["in_local_cache"]}\t{t["accession"]}\n')


def compile_databases():
    fasta_protein_db = Path(context.DATABASE_DIR, DB_PROTEINS_FILENAME)
    context.run_external(f'diamond makedb --in {fasta_protein_db} --db {fasta_protein_db}')
    fasta_nt_db = Path(context.DATABASE_DIR, DB_RNA_FILENAME)
    context.run_external(f'makeblastdb -in {fasta_nt_db} -dbtype nucl')

@context.register_database_installer
def format_blast_databases():
    if 'B' not in context.TASKS:
        return
    context.log('Concatenating and formatting blast databases for (prok, euk, vir)...')
    PROK_DB_DIR = Path(context.DATABASE_DIR, 'ncbi-cache', 'pro')
    EUK_DB_DIR = Path(context.DATABASE_DIR, 'ncbi-cache', 'euk')
    VIR_DB_DIR = Path(context.DATABASE_DIR, 'ncbi-cache', 'vir')

    for filename in DB_PROTEINS_FILENAME, DB_RNA_FILENAME, DB_DESCRIPTIONS_FILENAME, DB_TAXONOMY_FILENAME:
        with open(Path(context.DATABASE_DIR, filename), mode="wb") as destination:
            for dir in PROK_DB_DIR, EUK_DB_DIR, VIR_DB_DIR:
                src_file = Path(dir, filename)
                if src_file.exists():
                    with open(src_file, mode="rb") as source:
                        shutil.copyfileobj(source, destination)
    compile_databases()


import shutil

from metaerg import context
from metaerg.datatypes import sqlite

def _run_programs(genome_name, contig_dict, db_connection, result_files):
    cds_aa_file = context.spawn_file('cds.faa', genome_name)
    with open(result_files[0], 'w') as output, open(cds_aa_file) as input:
        context.run_external('tmhmm', stdin=input, stdout=output)


def _read_results(genome_name, contig_dict, db_connection, result_files) -> int:
    count = 0
    current_feature = None
    current_txt = ""
    feature_tmh_count = 0
    with open(result_files[0]) as tmhmm_handle:
        for line in tmhmm_handle:
            words = line.strip().split()
            match words:
                case[first_text, *_] if first_text.startswith('#'):
                    continue
                case [_, _, 'TMhelix', start, end]:
                    feature_tmh_count += 1
                    current_txt += f'{start}-{end},'
                case [next_feature_name, _, orientation, _, _] if orientation in ('inside', 'outside'):
                    next_feature = sqlite.read_feature_by_id(db_connection, next_feature_name)
                    if not next_feature:
                        raise Exception(f'Found results for unknown feature {next_feature_name}, '
                                        f'may need to rerun metaerg with --force')
                    if not current_feature or next_feature.id != current_feature.id:
                        if feature_tmh_count:
                            current_feature.tmh = feature_tmh_count
                            current_feature.tmh_topology = current_txt[:-1]
                            sqlite.update_feature_in_db(db_connection, current_feature)
                            count += 1
                        current_feature = next_feature
                        feature_tmh_count = 0
                        current_txt = 'i,' if 'inside' == orientation else 'o,'
        if feature_tmh_count:
            current_feature.tmh = feature_tmh_count
            current_feature.tmh_topology = current_txt[:-1]
            sqlite.update_feature_in_db(db_connection, current_feature)
            count += 1
    return count


@context.register_annotator
def run_and_read_tmhmm():
    return ({'pipeline_position': 111,
             'annotator_key': 'tmhmm',
             'purpose': 'transmembrane helix prediction with tmhmm',
             'programs': ('tmhmm',),
             'result_files': ('tmhmm',),
             'run': _run_programs,
             'read': _read_results})


def cleanup(dir):
    for file in dir.glob(f'TMHMM_*'):
        if file.is_dir():
            shutil.rmtree(file)



import math

import re
from time import sleep
from statistics import median, mean
from math import log, log10
from pathlib import Path
from collections import namedtuple

from tqdm import tqdm

from metaerg import context
from metaerg.datatypes.fasta import FastaParser
from metaerg.datatypes import ncbi_ftp
from metaerg.datatypes import sqlite

CUBData = namedtuple('CUBData', ['genome_id', 'nHE', 'CUBHE', 'consistency', 'CPB', 'filtered', 'd'])

def compute_feature_codon_counts_by_aa(feature, codon_counts_by_aa: dict|int = 0):
    if not isinstance(codon_counts_by_aa, dict):
        codon_counts_by_aa = {}

    # to deal with start codon, always translated as M:
    aa_seq = feature.aa_seq[0:1].lower() + feature.aa_seq[1:].upper()
    nt_seq = feature.nt_seq[0:3].lower() + feature.nt_seq[3:].upper()

    for i in range(1, len(aa_seq)):
        aa = aa_seq[i]
        codon = nt_seq[i * 3:i * 3 + 3]
        if 'N' in codon:
            continue
        if 'X' == aa:
            continue
        try:
            codons_for_aa = codon_counts_by_aa[aa]
        except KeyError:
            codons_for_aa = {}
            codon_counts_by_aa[aa] = codons_for_aa
        try:
            prev_count = codons_for_aa[codon]
        except:
            prev_count = 0
            codons_for_aa[codon] = 0
        codons_for_aa[codon] = prev_count + 1

    return codon_counts_by_aa


def convert_counts_to_frequencies(codon_counts_by_aa):
    codon_freq_by_aa = {}
    for aa, codons_for_aa in codon_counts_by_aa.items():
        codon_freq_by_aa[aa] = {}
        total_counts_for_aa = sum(codons_for_aa.values())
        for codon, count in codons_for_aa.items():
            codon_freq_by_aa[aa][codon] = count / total_counts_for_aa
    return codon_freq_by_aa


def compute_codon_frequencies_for_feature_data(db_connection, additional_sql: str = '') -> dict:
    genome_wide_codon_counts_by_aa = {}
    for feature in sqlite.read_all_features(db_connection, type='CDS', additional_sql=additional_sql):
        compute_feature_codon_counts_by_aa(feature, genome_wide_codon_counts_by_aa)
    genome_wide_codon_freq_by_aa = convert_counts_to_frequencies(genome_wide_codon_counts_by_aa)
    return genome_wide_codon_freq_by_aa


def compute_codon_usage_bias_for_feature(feature, genome_wide_codon_frequencies_by_aa) -> float:
    # see https://link.springer.com/article/10.1186/1471-2105-6-182, https://github.com/BioinfoHR/coRdon/blob/master/R/codonUsage.R, lines 218-255
    cds_codon_counts_by_aa = compute_feature_codon_counts_by_aa(feature)
    cds_codon_freq_by_aa = convert_counts_to_frequencies(cds_codon_counts_by_aa)
    milc = 0
    correction_factor = 0
    for aa, codons_for_aa in cds_codon_freq_by_aa.items():
        correction_factor += len(genome_wide_codon_frequencies_by_aa[aa]) - 1
        for codon, freq in codons_for_aa.items():
            milc += 2 * cds_codon_counts_by_aa[aa][codon] * log(freq / genome_wide_codon_frequencies_by_aa [aa][codon])
    milc /= len(feature.aa_seq) - 1
    correction_factor = correction_factor / (len(feature.aa_seq) - 1) - 0.5  # -0.5 as in coRdon, not +0.5 as in paper
    return milc - correction_factor


def compute_codon_pair_scores_for_feature_data(genome_id: str, db_connection, additional_sql: str = '') -> dict:
    # see https://www.science.org/doi/full/10.1126/science.1155761, supplement, Fig. 1
    codon_pair_counts = {}
    aa_pair_counts = {}
    aa_counts = {}
    codon_counts = {}
    codon2aa = {}

    ambiguous_codon_warnings = 0
    for feature in sqlite.read_all_features(db_connection, type='CDS', additional_sql=additional_sql):
        # to deal with start codon, always translated as M:
        aa_seq = feature.aa_seq[0:1].lower() + feature.aa_seq[1:].upper()
        nt_seq = feature.nt_seq[0:3].lower() + feature.nt_seq[3:].upper()

        for i in range(1, len(aa_seq)):
            x = aa_seq[i]
            a = nt_seq[i * 3:i * 3 + 3]
            if 'N' in a:
                continue
            if 'X' == x:
                continue
            try:
                prev_x = codon2aa[a]
                if x != prev_x:  # this very rarely happens
                    ambiguous_codon_warnings += 1
                    break
            except KeyError:
                pass
            codon2aa[a] = x
            try:
                aa_counts[x] += 1
            except KeyError:
                aa_counts[x] = 1
            try:
                codon_counts[a] += 1
            except KeyError:
                codon_counts[a] = 1
            try:
                y = aa_seq[i+1]
                b = nt_seq[(i + 1) * 3:(i + 1) * 3 + 3]
                if 'N' in b:
                    continue
                if 'X' == y:
                    continue
            except IndexError:
                break
            try:
                codon_pair_counts[a + b] += 1
            except KeyError:
                codon_pair_counts[a + b] = 1
            try:
                aa_pair_counts[x + y] += 1
            except KeyError:
                aa_pair_counts[x + y] = 1
    if ambiguous_codon_warnings > 0:
        print(
            f'Warning, while computing codon pair scores, skipped {ambiguous_codon_warnings}/{len(feature_data)} '
            f'orf with identical codon translated to different AAs in {genome_id}')

    codon_pair_scores = {}
    for ab, ab_counts in codon_pair_counts.items():
        a = ab[:3]
        b = ab[3:]
        x = codon2aa[a]
        y = codon2aa[b]
        codon_pair_scores[ab] = log(ab_counts / (codon_counts[a] * codon_counts[b] / (aa_counts[x] * aa_counts[y]) * aa_pair_counts[x+y]))
    return codon_pair_scores


def compute_codon_pair_bias_for_cds(feature, codon_pair_scores) -> float:
    # see https://www.science.org/doi/full/10.1126/science.1155761, supplement, Fig. 1
    cpb = 0
    nt_seq = feature.nt_seq[0:3].lower() + feature.nt_seq[3:].upper()
    for i in range(1, len(feature.aa_seq)):
        codon_pair = nt_seq[i * 3: i * 3 + 6]
        if 'N' in codon_pair:
            continue
        try:
            cpb += codon_pair_scores[codon_pair]
        except KeyError:
            pass  # can happen, a codon pair that was not seen before
    return cpb / len(feature.aa_seq)


def compute_codon_usage_bias_for_genome(genome_id, db_connection) -> CUBData:
    # codon usage bias
    background_frequencies = compute_codon_frequencies_for_feature_data(db_connection, additional_sql='end - start >= 240')
    cub_ribosomal = [compute_codon_usage_bias_for_feature(rp, background_frequencies)
                               for rp in sqlite.read_all_features(db_connection, type='CDS',
                               additional_sql='end - start >= 240 AND subsystems LIKE "%ribosomal protein%"')]
    if len(cub_ribosomal):
        codon_usage_bias = median(cub_ribosomal)
        # consistency in codon usage bias
        background_frequencies = compute_codon_frequencies_for_feature_data(db_connection,
                                 additional_sql='end - start >= 240 AND subsystems LIKE "%ribosomal protein%"')
        consistency = mean([compute_codon_usage_bias_for_feature(rp, background_frequencies)
                               for rp in sqlite.read_all_features(db_connection, type='CDS',
                               additional_sql='end - start >= 240 AND subsystems LIKE "%ribosomal protein%"')])
        # codon pair bias
        codon_pair_scores = compute_codon_pair_scores_for_feature_data(genome_id, db_connection, additional_sql='end - start >= 240')
        # codon_pair_bias = sum(codon_pair_scores.values()) / (len(codon_pair_scores) - 1)
        codon_pair_bias = median([compute_codon_pair_bias_for_cds(rp, codon_pair_scores)
                                  for rp in sqlite.read_all_features(type='CDS', additional_sql='end - start >= 240 AND subsystems LIKE "%ribosomal protein%"')])
        # keep track of # too short proteins filtered out and ribosomal proteins
        filtered_out = sum(1 for feature in sqlite.read_all_features(db_connection, type='CDS', additional_sql='end - start < 240'))
        return CUBData(genome_id, len(cub_ribosomal), codon_usage_bias, consistency, codon_pair_bias, filtered_out, 0)
    else:
        return CUBData(genome_id, 0, 0, 0, 0, 0, 0)


def compute_codon_bias_estimate_doubling_time(db_connection):
    background_frequencies = compute_codon_frequencies_for_feature_data(db_connection, additional_sql='end - start >= 240')
    cub_ribosomal = [compute_codon_usage_bias_for_feature(rp, background_frequencies)
                               for rp in sqlite.read_all_features(db_connection, type='CDS',
                               additional_sql='end - start >= 240 AND subsystems LIKE "%ribosomal protein%"')]
    if len(cub_ribosomal):
        codon_usage_bias = median(cub_ribosomal)
        doubling_time = math.pow(10, codon_usage_bias * -2.41937374 + 2.00361692)
        return codon_usage_bias, doubling_time
    else:
        context.log('Warning: No ribosomomal proteins for genome.')
        return 0, 0


def parse_training_data(list_file) -> list[CUBData]:
    training_data = []
    refseq_acc_pattern = re.compile(r'(GCF_\d+)')
    with open(list_file) as reader:
        for line in reader:
            words = line.split('\t')
            genome_id = words[-1]
            if m := refseq_acc_pattern.match(genome_id):
                genome_id = m.group(1)
                training_data.append(CUBData(genome_id, int(words[1]), float(words[2]), float(words[3]), float(words[4]),
                                        int(words[5]), float(words[6])))
    return training_data


def download_training_data(training_data: list[CUBData], destination_dir: Path):
    for entry in tqdm(training_data):
        accession = entry.genome_id
        targets = []
        extension = '_protein.faa.gz'
        aa_file = destination_dir / (accession + extension)
        if not aa_file.exists():
            targets.append((extension, aa_file))
        extension = '_cds_from_genomic.fna.gz'
        fna_file = destination_dir / (accession + extension)
        if not fna_file.exists():
            targets.append((extension, fna_file))
        if targets:
            ncbi_ftp.fetch(accession, targets)
            sleep(2)


def load_feature_data(dir: Path, entry: CUBData):
    nt_file = dir / (entry.genome_id + '_cds_from_genomic.fna.gz')
    aa_file = dir / (entry.genome_id + '_protein.faa.gz')
    aa_seq_dict = {}
    feature_data = []
    try:
        with FastaParser(aa_file, cleanup_seq=False) as reader:
            for seq in reader:
                aa_seq_dict[seq['id']] = seq

        pattern_prot_id = re.compile(r'\[protein_id=(.+?)]')
        with FastaParser(nt_file, cleanup_seq=False) as reader:
            for seq in reader:
                if m:= pattern_prot_id.search(seq['descr']):
                    prot_id = m.group(1)
                    feature_data.append({'id': prot_id,
                                         'subsystems': aa_seq_dict[prot_id]['descr'],
                                         'nt_seq': seq['seq'],
                                         'aa_seq': aa_seq_dict[prot_id]['seq'],
                                         'type': 'CDS'})
        if not feature_data:
            raise Exception(f'No data for {entry.genome_id}.')
        return pd.DataFrame(feature_data)
    except Exception:
        aa_file.unlink(missing_ok=True)
        nt_file.unlink(missing_ok=True)
        return None


def main():
    dir = Path('/bio/databases/eggo-data/')
    training_data = parse_training_data(dir / 'CodonStatistics_Training.csv')
    # download_training_data(training_data, Path('/bio/databases/eggo-data/genomes'))
    count = 0
    with open(dir / 'training_plus_metaerg.csv', 'w') as output:
        for entry in training_data:
            df = load_feature_data(dir / 'genomes', entry)
            if not df is None:
                metaerg_entry = compute_codon_usage_bias_for_genome(entry.genome_id, df)
                print(f'{count}/{len(training_data)} {metaerg_entry.nHE}/{entry.nHE}, '
                      f'{metaerg_entry.CUBHE:.3f}/{entry.CUBHE:.3f} '
                      f'{metaerg_entry.consistency:.3f}/{entry.consistency:.3f} '
                      f'{metaerg_entry.CPB:.3f}/{entry.CPB:.3f} '
                      f'{entry.genome_id}')
                count += 1
                output.write('{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\n'.format(entry.genome_id,
                                                                          entry.nHE,
                                                                          entry.filtered,
                                                                          entry.CUBHE,
                                                                          entry.consistency,
                                                                          entry.CPB,
                                                                          entry.d,
                                                                          metaerg_entry.nHE,
                                                                          metaerg_entry.CUBHE,
                                                                          metaerg_entry.consistency,
                                                                          metaerg_entry.CPB,
                                                                          ))


if __name__ == "__main__":
    main()


from metaerg import context
from metaerg.datatypes import fasta
from metaerg.datatypes import sqlite


def _run_programs(genome_name, contig_dict, db_connection, result_files):
    fasta_file = context.spawn_file('masked', genome_name)
    fasta.write_contigs_to_fasta(contig_dict, fasta_file, db_connection, genome_name,
                                 mask_targets=fasta.ALL_MASK_TARGETS)
    with open(result_files[0], 'w') as output:
        context.run_external(f'trf {fasta_file} 2 7 7 80 10 50 500 -d -h -ngs', stdout=output)

def _read_results(genome_name, contig_dict, db_connection, result_files) -> int:
    count = 0
    with open(result_files[0]) as trf_handle:
        for line in trf_handle:
            if line.startswith("@"):
                contig_name = line[1:].strip()
                if space_index := contig_name.find(' ') > 0:
                    contig_name = contig_name[:space_index]
                contig = contig_dict[contig_name]
                continue
            if not contig:
                continue
            words = line.split()
            start = int(words[0]) - 1
            end = int(words[1])
            seq = contig['seq'][start:end]
            feature = sqlite.Feature(genome = genome_name,
                                     contig = contig['id'],
                                     start = start,
                                     end = end,
                                     strand = 1,
                                     type = 'repeat',
                                     inference = 'tandem-repeat-finder',
                                     nt_seq = seq,
                                     notes = f'period size {words[2]}; copies {words[3]}')
            sqlite.add_new_feature_to_db(db_connection, feature)
            count += 1
    return count

@context.register_annotator
def run_and_read_trf():
    return ({'pipeline_position': 41,
             'annotator_key': 'trf',
             'purpose': 'tandem repeat prediction with trf',
             'programs': ('trf',),
             'result_files': ('tandem-repeat-finder',),
             'run': _run_programs,
             'read': _read_results})



from pathlib import Path
from metaerg import context

GENOME_PROPERTY_FORMATS = {'genome name': '<',
                           'input fasta file': '<',
                           '# contigs': ',',
                           'size': ',',
                           '% GC': '.1%',
                           'N50': ',',
                           '# proteins': ',',
                           '% coding': '.1%',
                           'mean protein length (aa)': '.1f',
                           '# ribosomal RNA': ',',
                           '# transfer RNA': ',',
                           '# non-coding RNA': ',',
                           '# retrotransposons': ',',
                           '# CRISPR repeats': ',',
                           '# other repeats': ',',
                           '# total features': ',',
                           '% repeats': '.1%',
                           '% CDS classified to taxon': '.1%',
                           '% of CDS classified to dominant taxon': '.1%',
                           'dominant taxon': '<',
                           'codon usage bias': '.3f',
                           'doubling_time (days)': '.1f'
                           }

@context.register_html_writer
def write_html(genome_name, db_connection, genome_properties:dict, dir):
    dir.mkdir(exist_ok=True, parents=True)
    file = Path(dir, genome_name, 'index.html')
    with open(file, 'w') as handle:
        handle.write(make_html(genome_name, genome_properties))


def make_html(genome_name, genome_properties:dict) -> str:
    """injects the content into the html base, returns the html"""
    html = _make_html_template()
    html = html.replace('GENOME_NAME', genome_name)
    # genome properties
    html = html.replace('CONTENT_PROPERTIES', ''.join((f'<tr><td>{k}</td><td>{genome_properties[k]:{GENOME_PROPERTY_FORMATS[k]}}</td></tr>\n'
                                                       for k in GENOME_PROPERTY_FORMATS.keys())))
    subsystem_html = ''
    subsystem_data = genome_properties['subsystems']
    # subsystem_data
    for subsystem in subsystem_data.index.unique('subsystem'):
        sub_data = subsystem_data.loc[subsystem, :]
        subsystem_html += f'<b id=f>{subsystem}</b>'
        if 'secondary-metabolites' == subsystem:
            if len(sub_data.index):
                subsystem_html += '<p id=f><a href="antismash/index.html" target="">View antismash results.</a></p>\n'
        else:
            subsystem_html += '<table id=f>\n'
            for function in sub_data.itertuples():
                if function.Index:
                    subsystem_html += f'<tr><td>{function.Index}</td><td>\n'
                    subsystem_html += ', '.join('<a target="Gene Details" href="feature-details.html#{}">{}</a>'
                                                .format(f_id.split("@")[0], f_id) for f_id in function.genes.split())
                subsystem_html += f'</td></tr>\n'
            subsystem_html += '</table>\n'
    html = html.replace('CONTENT_SUBSYSTEMS', subsystem_html)
    return html


def _make_html_template() -> str:
    """should return the html base for injecting the content in. Returns the html"""
    return '''<!DOCTYPE html>
<html>
<head>
    <meta charset="utf-8">
        <title>GENOME_NAME - properties and subsystems</title>
</head>
<body>

<style>
  th {
    background-color: white;
     }
  #f {
    font-family: Calibri, sans-serif;
    text-align: left;
    padding: 0px;
    margin: 0px;
     }
  .accordion {
    background-color: #eee;
    color: #444;
    cursor: pointer;
    padding: 10px;
    width: 100%;
    border: none;
    text-align: left;
    outline: none;
    font-size: 15px;
    transition: 0.4s;
  }
  .active, .accordion:hover {
    background-color: #ccc; 
  }
  .panel {
    font-family: Calibri, sans-serif;
    padding: 0 18px;
    display: none;
    background-color: white;
    overflow: hidden;
  }
</style>

<h3 id=f>Genome properties</h4>
<table id=f>
    <thead>
        <tr>
          <th>property</th>
          <th>value</th>
        </tr>
    </thead>
    <tbody>
CONTENT_PROPERTIES
    </tbody></table>
    <p></p>
    <h4 id=f><a href="feature_table.html">View table with all genes</a></h4>
    <p></p>
    <h3 id=f>Subsystems overview</h3>
CONTENT_SUBSYSTEMS
</body></html>'''



import shutil
from pathlib import Path
from concurrent.futures import ProcessPoolExecutor
from collections import namedtuple

from metaerg import context
from metaerg.datatypes import fasta
from metaerg.datatypes import sqlite


def _run_programs(genome_name, contig_dict, db_connection, result_files):
    fasta_file = context.spawn_file('masked', genome_name)
    rfam_database = Path(context.DATABASE_DIR, 'rfam', 'Rfam.cm')
    if context.CPUS_PER_GENOME > 1:
        split_fasta_files = fasta.write_contigs_to_fasta(contig_dict, fasta_file, db_connection, genome_name,
                                                         mask_targets=fasta.ALL_MASK_TARGETS,
                                                         split=context.CPUS_PER_GENOME)
        split_cmscan_files = [Path(result_files[0].parent, f'{result_files[0].name}.{i}')
                              for i in range(len(split_fasta_files))]
        with ProcessPoolExecutor(max_workers=context.CPUS_PER_GENOME) as executor:
            for split_input, split_output in zip(split_fasta_files, split_cmscan_files):
                executor.submit(context.run_external, f'cmscan --rfam --tblout {split_output} {rfam_database} '
                                                      f'{split_input}')
        with open(result_files[0], 'wb') as output:
            for split_input_file, split_output_file in zip(split_fasta_files, split_cmscan_files):
                with open(split_output_file, 'rb') as input:
                    shutil.copyfileobj(input, output)
                split_input_file.unlink()
                split_output_file.unlink()
    else:
        fasta.write_contigs_to_fasta(contig_dict, fasta_file, db_connection, genome_name,
                                     mask_targets=fasta.ALL_MASK_TARGETS)
        context.run_external(f'cmscan --rfam --tblout {result_files[0]} {rfam_database} {fasta_file}')


def _read_results(genome_name, contig_dict, db_connection, result_files) -> int:
    NON_CODING_RNA_TYPES = {'LSU_rRNA_bacteria': 'rRNA',
                            'LSU_rRNA_archaea': 'rRNA',
                            'LSU_rRNA_eukarya': 'rRNA',
                            'SSU_rRNA_bacteria': 'rRNA',
                            'SSU_rRNA_archaea': 'rRNA',
                            'SSU_rRNA_eukarya': 'rRNA',
                            'SSU_rRNA_microsporidia': 'rRNA',
                            '5S_rRNA': 'rRNA',
                            '5_8S_rRNA': 'rRNA',
                            'tmRNA': 'tmRNA',
                            'tRNA': 'tRNA'}
    hits = []
    Hit = namedtuple('Hit', ('query_id', 'hit_id', 'query_start', 'query_end',
                             'query_strand', 'score', 'descr'))
    with open(result_files[0]) as hmm_handle:
        for line in hmm_handle:
            words = line.strip().split()
            if line.startswith('#') or len(words) < 17:
                continue
            words[17] = ' '.join(words[17:])
            match words:
                case [hit, _, query, _, _, _, _, start, end, '-', _, _, _, _, score, _, '!', descr, *_]:
                    hit = Hit(query, hit, int(end), int(start), -1, float(score), descr)
                case [hit, _, query, _, _, _, _, start, end, '+', _, _, _, _, score, _, '!', descr, *_]:
                    hit = Hit(query, hit, int(start), int(end), 1, float(score), descr)
                case [*_]:
                    continue
            overlap = None
            for prev_hit in hits:
                if hit.query_id == prev_hit.query_id and hit.query_start < prev_hit.query_end and \
                        hit.query_end > prev_hit.query_start:
                    overlap = prev_hit  # overlap detected
                    break
            if overlap:
                if hit.score > overlap.score:
                    hits.remove(overlap)
                    hits.append(hit)
            else:
                hits.append(hit)
    count = 0
    for hit in hits:
        if hit.hit_id in NON_CODING_RNA_TYPES.keys():
            f_type = NON_CODING_RNA_TYPES[hit.hit_id]
        elif hit.hit_id.startswith('CRISPR'):
            f_type = 'crispr_repeat'
        else:
            f_type = 'ncRNA'
        contig = contig_dict[hit.query_id]
        seq = contig['seq'][hit.query_start - 1:hit.query_end]
        if hit.query_strand < 0:
            seq = fasta.reverse_complement(seq)
        feature = sqlite.Feature(genome = genome_name,
                   contig = hit.query_id,
                   start = hit.query_start - 1,
                   end = hit.query_end,
                   strand = hit.query_strand,
                   type = f_type,
                   inference = 'cmscan',
                   nt_seq = seq,
                   descr = "{} {}".format(hit.hit_id, hit.descr))
        sqlite.add_new_feature_to_db(db_connection, feature)
        count += 1
    return count


@context.register_annotator
def run_and_read_cmscan():
    return ({'pipeline_position': 21,
             'annotator_key': 'cmscan',
             'purpose': 'noncoding (RNA) gene prediction with cmscan',
             'programs': ('cmscan',),
             'result_files': ("cmscan",),
             'databases': (Path('rfam', 'Rfam.cm'),),
             'run': _run_programs,
             'read': _read_results})


@context.register_database_installer
def install_cmscan_database():
    if 'R' not in context.TASKS:
        return
    rfam_dir = Path(context.DATABASE_DIR, 'rfam')
    rfam_dir.mkdir(exist_ok=True, parents=True)

    rfam_file = Path(rfam_dir, 'Rfam.cm')
    if context.FORCE or not rfam_file.exists() or not rfam_file.stat().st_size:
        context.log(f'Installing the RFAM database to {rfam_file}...')
        context.run_external(
            f'wget -P {rfam_dir} http://ftp.ebi.ac.uk/pub/databases/Rfam/CURRENT/Rfam.cm.gz')
        context.run_external(f'gunzip {rfam_file}.gz')
    else:
        context.log(f'Keeping existing conserved domain database in {rfam_file}, use --force to overwrite.')
    if context.FORCE or not Path(context.DATABASE_DIR, "Rfam.cm.i1f").exists():
        context.log(f'Running cmpress...')
        context.run_external(f'cmpress -F {rfam_file}')
    else:
        context.log('Skipping cmpress for previously cmpressed RFAM database...')



from urllib.parse import unquote
import re
import gzip
from metaerg.datatypes import fasta
from metaerg.datatypes.sqlite import Feature

class GffParser:
    def __init__(self, path, contig_dict, target_feature_type_dict:dict=None, inference:str=None):
        self.path = path
        self.contig_dict = contig_dict
        self.target_feature_type_dict = target_feature_type_dict  # should be a dictionary to convert feature types
        self.inference = inference

    def __enter__(self):
        if str(self.path).endswith('.gz'):
            self.handle = gzip.open(self.path, 'rt')
        else:
            self.handle = open(self.path)
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.handle.close()

    def __iter__(self):
        while line := self.handle.readline():
            words = line.strip().split('\t')
            match(words):
                case[word, *_] if word.startswith('#'):
                    continue
                case [contig_name, inference, feature_type, start, end, score, strand, frame, qualifiers]:
                    if self.target_feature_type_dict and feature_type not in self.target_feature_type_dict.keys():
                        continue
                    try:
                        contig = self.contig_dict[contig_name]
                    except KeyError:
                        continue
                    start = int(start) - 1
                    end = int(end)
                    strand = -1 if '-' == strand else 1
                    seq = contig['seq'][start:end]
                    if strand < 0:
                        seq = fasta.reverse_complement(seq)
                    inference = self.inference if self.inference else inference
                    qualifiers = re.split(r"[=;]", qualifiers)
                    qualifiers = [unquote(str) for str in qualifiers]
                    if len(qualifiers) % 2 != 0:
                        qualifiers = qualifiers[:-1]  # this happens for example with prodigal, ending with ";"
                    qualifiers = {qualifiers[i].lower(): qualifiers[i + 1] for i in range(0, len(qualifiers), 2)}
                    feature = Feature(contig=contig_name,
                               start=start,
                               end=end,
                               strand=strand,
                               type=self.target_feature_type_dict[feature_type],
                               inference=inference,
                               nt_seq=seq)
                    yield feature


def parse_feature_qualifiers_from_gff(qualifier_str) -> {}:
    qal = re.split(r"[=;]", qualifier_str)
    qal = [unquote(str) for str in qal]
    if len(qal) % 2 != 0:
        qal = qal[:-1]  # this happens for example with prodigal which has the qualifier column ending with ";"
    qal = {qal[i].lower(): qal[i + 1] for i in range(0, len(qal), 2)}
    return qal



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name='metaerg',
    version='2.3.19',
    packages=setuptools.find_packages(where='src'),
    url='https://github.com/kinestetika/MetaErg',
    license='MIT',
    author='Marc Strous',
    author_email='mstrous@ucalgary.ca',
    description='Annotation of genomes and contigs',
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=['Development Status :: 4 - Beta',
                 'Environment :: Console',
                 'Natural Language :: English',
                 'Operating System :: OS Independent',
                 'License :: OSI Approved :: MIT License',
                 'Programming Language :: Python :: 3.10',
                 'Topic :: Scientific/Engineering :: Bio-Informatics'],
    keywords='repeat-regions genes functions taxonomy',
    project_urls={'Source': 'https://github.com/kinestetika/MetaErg'},
    package_dir={'': 'src'},
    package_data={'': ['functional_gene_data']},
    python_requires='>=3.10',
    install_requires=['biopython', 'ncbi-datasets-pylib', 'pandas', 'httpx', 'virtualenv', 'h5py', 'pyarrow', 'openpyxl', 'tqdm'],
    extras_require={  # Optional
        'dev': ['setuptools', 'build', 'twine'],
        'test': []
    },
    entry_points={  # Optional
        'console_scripts': [
            'metaerg=metaerg.main:main',
        ],
    }
)


from metaerg import context
from metaerg.datatypes import fasta, gff, sqlite


def _run_programs(genome_name, contig_dict, db_connection, result_files):
    """Executes the helper programs to complete the analysis"""
    fasta_file = context.spawn_file('masked', genome_name)
    fasta.write_contigs_to_fasta(contig_dict, fasta_file, db_connection, genome_name,
                                 mask_targets=fasta.ALL_MASK_TARGETS)
    context.run_external(f'minced -gffFull {fasta_file} {result_files[0]}')


def _read_results(genome_name, contig_dict, db_connection, result_files) -> int:
    """Should parse the result files and return the # of positives"""
    with gff.GffParser(result_files[0], contig_dict, inference='minced',
                       target_feature_type_dict={'repeat_unit': 'crispr_repeat'}) as gff_parser:
        count = 0
        for feature in gff_parser:
            feature.genome = genome_name
            sqlite.add_new_feature_to_db(db_connection, feature)
            count += 1
    return count


@context.register_annotator
def run_and_read_minced():
    return ({'pipeline_position': 1,
             'annotator_key': 'minced',
             'purpose': 'CRISPR prediction with minced',
             'programs': ('minced',),
             'result_files': ("minced",),
             'run': _run_programs,
             'read': _read_results})




from ftplib import FTP, error_perm

NCBI_FTP = None


def fetch(accession, targets: list) -> list[bool]:
    # targets is a list of tuples (extension, destination_file)
    global NCBI_FTP
    if not NCBI_FTP:
        NCBI_FTP = FTP('ftp.ncbi.nlm.nih.gov')
        NCBI_FTP.login()

    outcomes = []
    try:
        NCBI_FTP.cwd('/genomes/all/')
        for acc_part in (accession[0:3], accession[4:7], accession[7:10], accession[10:13]):
            NCBI_FTP.cwd(acc_part)
        ftp_dir_list = []
        NCBI_FTP.dir('.', ftp_dir_list.append)
        NCBI_FTP.cwd(ftp_dir_list[0].split()[-1])
        ftp_dir_list = []
        NCBI_FTP.dir('.', ftp_dir_list.append)
        for extension, destination_file in targets:
            success = False
            for l in ftp_dir_list:
                filename = l.split()[-1]
                if filename.endswith(extension):
                    success = True
                    with open(destination_file, "wb") as local_handle:
                        NCBI_FTP.retrbinary("RETR " + filename, local_handle.write)
            outcomes.append(success)
        return outcomes
    except EOFError:
        print('FTP Error at NCBI - resetting connection')
        NCBI_FTP = FTP('ftp.ncbi.nlm.nih.gov')
        NCBI_FTP.login()
    except BrokenPipeError:
        print('FTP Error at NCBI - resetting connection')
        NCBI_FTP = FTP('ftp.ncbi.nlm.nih.gov')
        NCBI_FTP.login()
    except error_perm:
        print(f'Unable to find {accession}')
        return outcomes